// Copyright (C) 1996 DIMACS Center, Rutgers, The State University of New Jersey
// Author(s): Patricia K. Fasel (Los Alamos Nat. Lab.), Jonathan Berry

// This software is copyrighted by the DIMACS Center at Rutgers, The State
// University of New Jersey.  IT IS PROVIDED AS IS, AND THE AUTHORS, DIMACS, AND
// RUTGERS, THE STATE UNIVERSITY OF NEW JERSEY  DISCLAIM
// ALL LIABILITY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
// DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
// DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

// THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
// IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
// NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
// MODIFICATIONS.

// The authors hereby grant permission to use, copy, modify, distribute,
// and license this software and its documentation for any purpose, provided
// that existing copyright notices are retained in all copies and that this
// notice is included verbatim in any distributions. No written agreement,
// license, or royalty fee is required for any of the authorized uses.
// Modifications to this software may be copyrighted by their authors
// and need not follow the licensing terms described here, provided that
// the new terms are clearly indicated on the first page of each file where
// they apply.

// Last File Update: 31-Jul-1996
// 

#define Container_CC


template <class Item>
void
SimpleContainer<Item>::concatenate(const Container<Item>& c)
{
    Iterator<Item> get_next(&c);    
    Item item;
    while (get_next(item))
        append(item);
}

template <class Item>
Item
Container<Item>::info(const ContainerNode *cn) const
{
	if (cn != 0)
		return  * (Item *) cn->info();
	static Item i;
	return i;
}

template <class Item>
Item
Container<Item>::max() const
{
    Iterator<Item> get_next(this);
    Item i, max_item;
    Bool start = FALSE;
    while (get_next(i))
        if (!start) {
                max_item = i;
                start = TRUE;
        } else {
		if (ElementOps<Item>::compareItems(i, max_item) > 0)
			max_item = i;
        }
    return max_item;
}

template <class Item>
Item
Container<Item>::min() const
{
    Iterator<Item> get_next(this);
    Item i, min_item;
    Bool start = FALSE;
    while (get_next(i))
        if (!start) {
                min_item = i;
                start = TRUE;
        } else {
		if (ElementOps<Item>::compareItems(i, min_item) < 0)
			min_item = i;
        }
    return min_item;
}
		

////*******************************************************************
template <class Item>
Bool
Container<Item>::operator==(const Container<Item>& m) const
{
    	Iterator<Item> get_next1(this);
    	Iterator<Item> get_next2(&m);
    	Item item1;
    	Item item2;
	int i;

	while ( (i=get_next1(item1)) && get_next2(item2)) {
		if (ElementOps<Item>::compareItems(item1,item2)!=0)
			return(FALSE);
	}
	if (i || get_next2(item2))      // not same size
		return(FALSE);		// note that an element count
	return(TRUE);			// is not stored (to allow very  
}					// close ties with STk.  Therefore
					// the size() query is O(n) and
					// so we do not use it here.

template <class Item>
Bool
Container<Item>::operator>(const Container<Item>& m) const
{
    Iterator<Item> get_next1(this);
    Iterator<Item> get_next2(&m);
    Item item1;
    Item item2;

    while (get_next1(item1))
    {
	if (get_next2(item2))
	{
	    if (ElementOps<Item>::compareItems(item1, item2)>0)
	   	 return(TRUE);
	    if (ElementOps<Item>::compareItems(item1, item2)<0)
	       	 return(FALSE);
	}
	else return(TRUE);
    }
    return(FALSE);
}

template <class Item>
Bool
Container<Item>::operator>=(const Container<Item>& m) const
{
    return !operator<(m);
}

template <class Item>
Bool
Container<Item>::operator<(const Container<Item>& m) const
{
    Iterator<Item> get_next1(this);
    Iterator<Item> get_next2(&m);
    Item item1;
    Item item2;

    while (get_next1(item1))
    {
        if (get_next2(item2))
	{
	    if (ElementOps<Item>::compareItems(item1, item2)<0)
                return(TRUE);
	    if (ElementOps<Item>::compareItems(item1, item2)>0)
	        return(FALSE);
	}
	else return(FALSE);
    }
    if (get_next2(item2))
    	return(TRUE);
    return(FALSE);
}

template <class Item>
Bool
Container<Item>::operator<=(const Container<Item>& m) const
{
    return !operator>(m);
}

template <class Item>
Bool
Container<Item>::operator!=(const Container<Item>& m) const
{
    return !operator==(m);
}

//*******************************************************************



template <class Item>
Bool
Container<Item>::operator==(const Collection<Item>& m) const
{
    	Iterator<Item> get_next1(this);
    	Iterator<Item> get_next2(&m);
    	Item item1;
    	Item item2;
	int i;

	while ( (i=get_next1(item1)) && get_next2(item2)) {
		if (ElementOps<Item>::compareItems(item1,item2)!=0)
			return(FALSE);
	}
	if (i || get_next2(item2))      // not same size
		return(FALSE);		// note that an element count
	return(TRUE);			// is not stored (to allow very  
}					// close ties with STk.  Therefore
					// the size() query is O(n) and
					// so we do not use it here.

template <class Item>
Bool
Container<Item>::operator>(const Collection<Item>& m) const
{
    Iterator<Item> get_next1(this);
    Iterator<Item> get_next2(&m);
    Item item1;
    Item item2;

    while (get_next1(item1))
    {
	if (get_next2(item2))
	{
	    if (ElementOps<Item>::compareItems(item1, item2)>0)
	   	 return(TRUE);
	    if (ElementOps<Item>::compareItems(item1, item2)<0)
	       	 return(FALSE);
	}
	else return(TRUE);
    }
    return(FALSE);
}

template <class Item>
Bool
Container<Item>::operator>=(const Collection<Item>& m) const
{
    return !operator<(m);
}

template <class Item>
Bool
Container<Item>::operator<(const Collection<Item>& m) const
{
    Iterator<Item> get_next1(this);
    Iterator<Item> get_next2(&m);
    Item item1;
    Item item2;

    while (get_next1(item1))
    {
        if (get_next2(item2))
	{
	    if (ElementOps<Item>::compareItems(item1, item2)<0){
		return(TRUE);
	    }
	    if (ElementOps<Item>::compareItems(item1, item2)>0) {
		return(FALSE);
	    }
	}
	else {
	    return(FALSE);
	}
    }
    if (get_next2(item2)) {
	return(TRUE);
    }
    return(FALSE);
}

template <class Item>
Bool
Container<Item>::operator<=(const Collection<Item>& m) const
{
    return !operator>(m);
}

template <class Item>
Bool
Container<Item>::operator!=(const Collection<Item>& m) const
{
    return !operator==(m);
}

template <class Item> ostream& operator<<(ostream& os, const Container<Item>&c)
{
        c.display(os); return os;
}

